import configparser
import json
import os
import socket
from enum import Enum
from itertools import cycle
from pathlib import Path

import requests
import yaml


class RunCode(Enum):
    CONTINUOUS = 0
    ONCE = 1
    UNTIL_SUCCESS = 2
    MONTHLY = 3


class Account:

    def __init__(self, profile):
        try:
            parser = configparser.ConfigParser()
            parser.read(os.path.join(Path.home(), '.prelude', 'keychain.ini'))
            profile = next(s for s in parser.sections() if s == profile)
            self.hq = parser.get(profile, 'hq')
            self.key = parser.get(profile, 'key')
            self.headers = dict(account=parser.get(profile, 'account'), token=parser.get(profile, 'token'))
        except FileNotFoundError:
            raise Exception('Please create a ~/.prelude/keychain.ini file')
        except KeyError as e:
            raise Exception('Property not found, %s' % e)

    def details(self):
        """ Get Account object """
        res = requests.get('%s/account' % self.hq, headers=self.headers)
        if res.status_code == 200:
            return res.json()
        raise Exception('Failed to get Account (reason:%s)' % res.status_code)

    def change_token(self, token: str):
        """ Change Account token """
        requests.put('%s/account' % self.hq, headers=self.headers, json=dict(token=token))

    def add_outpost(self, address: str, token: str):
        """ Add or replace an Outpost server """
        credentials = dict(outpost_address=address, outpost_token=token)
        requests.put('%s/account' % self.hq, headers=self.headers, json=credentials)

    def register(self, endpoint=socket.gethostname()):
        """ Register a new endpoint to your account """
        payload = dict(identifiers=[endpoint])
        res = requests.post('%s/account/endpoint' % self.hq, headers=self.headers, json=payload)
        if res.status_code == 200:
            return res.json().get(endpoint)
        raise Exception('Failed to register endpoint: %s' % res.text)

    def search(self, ttp=None):
        """ Get TTPs matching your query """
        res = requests.get('%s/account/ttp' % self.hq, headers=self.headers, params=dict(id=ttp))
        if res.status_code == 200:
            return json.loads(self._xor(res.text))
        elif res.status_code == 503:
            raise Exception(res.text)
        raise Exception('Failed to locate %s. Do you own it?' % ttp)

    def upload(self, path: str):
        """ Save a TTP on your Outpost server """
        with open(path, 'r') as stream:
            procedure = None
            if path.endswith('.yml'):
                procedure = yaml.safe_load(stream)
            procedure = procedure if procedure else json.loads(stream.read())

            encrypted_ttp = self._xor(json.dumps(procedure))
            res = requests.post('%s/account/ttp' % self.hq, headers=self.headers, data=encrypted_ttp)
            if res.status_code != 200:
                raise Exception('Failed to upload: %s' % res.text)

    def activate(self, ttp: str, run_code=RunCode.ONCE):
        """ Activate a TTP so endpoints will start running it """
        res = requests.post(
            url='%s/account/activation/%s' % (self.hq, ttp),
            headers=self.headers,
            json=dict(run_code=run_code.value)
        )
        if res.status_code != 200:
            raise Exception('Failed to activate: %s' % res.text)

    def deactivate(self, ttp: str):
        """ Deactivate a TTP so endpoints will stop running it """
        res = requests.delete('%s/account/activation/%s' % (self.hq, ttp), headers=self.headers)
        if res.status_code != 200:
            raise Exception('Failed to deactivate: %s' % res.text)

    def transfer(self, ttp: str):
        """ Copy Prelude TTPs into Outpost """
        res = requests.get('%s/ttp' % self.hq, headers=self.headers, params=dict(id=ttp))
        if res.status_code == 200:
            encrypted_ttp = self._xor(json.dumps(res.json()))
            res2 = requests.post('%s/account/ttp' % self.hq, headers=self.headers, data=encrypted_ttp)
            if res2.status_code != 200:
                print('Failed to transfer: %s, %s' % (ttp, res2.text))
            print('Completed transfer of %s' % ttp)
        elif res.status_code == 503:
            raise Exception(res.text)
        else:
            raise Exception('Failed to sync procedures: %s' % res.text)

    def results(self, endpoint_token=None, days=7):
        """ Get a summary of Account activity, or an individual Endpoint """
        params = dict(token=endpoint_token, days=days)
        res = requests.get('%s/account/results' % self.hq, headers=self.headers, params=params)
        if res.status_code == 200:
            return res.json()
        raise Exception('Failed to get procedure: %s' % res.text)

    """ PRIVATE """

    def _xor(self, msg):
        return ''.join(chr(ord(c) ^ ord(k)) for c, k in zip(msg, cycle(self.key)))
